/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTION_POINTS;
import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_ANGLE;
import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_OFFSET;
import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION;
import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_POSITION;
import static org.fortiss.tooling.base.model.layout.EOrientation.NORTH;
import static org.fortiss.tooling.base.utils.AngleUtils.getAngle;
import static org.fortiss.tooling.base.utils.AngleUtils.setAngle;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodeSize;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createAngle;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createDimension;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createOffsetOrientation;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoint;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoints;
import static org.fortiss.tooling.base.utils.OffsetOrientationUtils.getOffsetOrientation;
import static org.fortiss.tooling.base.utils.OffsetOrientationUtils.setOffsetOrientation;
import static org.fortiss.tooling.base.utils.PointUtils.getPosition;
import static org.fortiss.tooling.base.utils.PointUtils.setPoint;
import static org.fortiss.tooling.base.utils.PointsUtils.getPoints;
import static org.fortiss.tooling.base.utils.PointsUtils.setPoints;

import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.layout.LayoutKeyConstants;
import org.fortiss.tooling.base.model.layout.Angle;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.LayoutFactory;
import org.fortiss.tooling.base.model.layout.Offset;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.utils.LayoutDataUtils;

/**
 * Utility methods for {@link ILayoutedModelElement} and {@link ILayoutData}.
 * This class provides access to layout data using the default keys defined in
 * {@link LayoutKeyConstants}. Changing some layout data is done by replacing
 * the layout data object under a given key, which again results in a EMF
 * notification on the {@link ILayoutedModelElement}. Therefore a single change
 * notification listener on the layouted element can be used.
 * 
 * @author hoelzl
 * @author wandinger
 * @author hummel
 */
public class LayoutDataUIUtils {

	/** Returns the bounds rectangle of the given layouted element. */
	public static Rectangle getNodeBounds(ILayoutedModelElement layouted) {
		org.fortiss.tooling.base.model.layout.Rectangle rectElement =
				LayoutDataUtils.getNodeBounds(layouted);
		return new Rectangle(rectElement.getX(), rectElement.getY(), rectElement.getWidth(),
				rectElement.getHeight());
	}

	/** Moves and resizes the given node. */
	public static void moveAndResizeNode(ILayoutedModelElement layouted, int dx, int dy, int dw,
			int dh, int minWidth, int minHeight) {
		Point pos = getNodePosition(layouted);
		int posX = pos.getX() + dx;
		int posY = pos.getY() + dy;

		Dimension dim = getNodeSize(layouted);
		int dimW = dim.getWidth() + dw;
		int dimH = dim.getHeight() + dh;

		if(dimW < minWidth) {
			if(dx > 0) {
				posX = posX - minWidth - dimW;
			}
			dimW = minWidth;
		}

		if(dimH < minHeight) {
			if(dy > 0) {
				posY = posY - minHeight - dimH;
			}
			dimH = minHeight;
		}

		setNodePosition(layouted, posX, posY);
		setNodeSize(layouted, dimW, dimH);
	}

	/** Returns the layouted connector's position. */
	public static Point getConnectorPosition(ILayoutedModelElement layouted) {
		return getPosition(layouted, CONNECTOR_POSITION);
	}

	/** Sets the layouted connector's position. */
	public static void setConnectorPosition(ILayoutedModelElement layouted, int x, int y) {
		setPoint(layouted, CONNECTOR_POSITION, x, y);
	}

	/** Returns the layouted connector's offset-orientation. */
	public static OffsetOrientation getConnectorOffsetOrientation(ILayoutedModelElement layouted) {
		OffsetOrientation oo = getOffsetOrientation(layouted, CONNECTOR_OFFSET_ORIENTATION);
		if(oo == null) {
			return createOffsetOrientation(0, NORTH, CONNECTOR_OFFSET_ORIENTATION);
		}
		return oo;
	}

	/** Returns the layouted connector's orientation. */
	public static EOrientation getConnectorOrientation(ILayoutedModelElement layouted) {
		return getOffsetOrientation(layouted, CONNECTOR_OFFSET_ORIENTATION).getOrientation();
	}

	/** Sets the layouted connector's offset and orientation. */
	public static void setConnectorOffsetOrientation(ILayoutedModelElement layouted,
			OffsetOrientation orientation) {
		setOffsetOrientation(layouted, CONNECTOR_OFFSET_ORIENTATION, orientation.getOffset(),
				orientation.getOrientation());
	}

	/** Returns the layouted connector's offset. */
	public static Offset getConnectorOffset(ILayoutedModelElement layouted) {
		OffsetOrientation offset = getOffsetOrientation(layouted, CONNECTOR_OFFSET);
		if(offset == null) {
			return createOffsetOrientation(0, NORTH, CONNECTOR_OFFSET);
		}
		return offset;
	}

	/** Sets the layouted connector's angle. */
	public static void setConnectorAngle(ILayoutedModelElement layouted, double angle) {
		setAngle(layouted, CONNECTOR_ANGLE, angle);
	}

	/** Returns the layouted connector's angle. */
	public static Angle getConnectorAngle(ILayoutedModelElement layouted) {
		Angle angle = getAngle(layouted, CONNECTOR_ANGLE);
		if(angle == null) {
			return createAngle(0.0, CONNECTOR_ANGLE);
		}
		return angle;
	}

	/** Returns the angle as double value (0 if not existing). */
	public static double getConnectorAngleAsDouble(ILayoutedModelElement lobject) {
		return getConnectorAngle(lobject).getAngle();
	}

	/**
	 * Return the absolute position for a given offset, orientation, size of the
	 * parent and insets.
	 */
	public static Point getAbsolute(OffsetOrientation oo, Rectangle parentBounds,
			int parentInsets) {
		Dimension d = createDimension(parentBounds.width, parentBounds.height, "");
		return getAbsolute(oo, d, parentInsets);
	}

	/**
	 * Return the absolute position for a given offset, orientation, size of the
	 * parent and insets.
	 */
	public static Point getAbsolute(OffsetOrientation oo, Dimension parentSize, int parentInsets) {
		Point pos = createPoint(0, 0, null);
		int offsetValue = oo.getOffset();
		switch(oo.getOrientation()) {
			case NORTH:
				pos.setX(offsetValue);
				pos.setY(0);
				break;
			case WEST:
				pos.setX(0);
				pos.setY(offsetValue);
				break;
			case SOUTH:
				pos.setX(offsetValue);
				pos.setY(parentSize.getHeight() - 2 * parentInsets);
				break;
			case EAST:
				pos.setX(parentSize.getWidth() - 2 * parentInsets);
				pos.setY(offsetValue);
				break;
		}
		return pos;
	}

	/** Returns the {@link Points} layout data object for the given connection. */
	public static Points getConnectionPoints(ILayoutedModelElement lobject) {
		return getPoints(lobject, CONNECTION_POINTS);
	}

	/** Adds the given x/y point to the layout data of the given element. */
	public static void addConnectionPoint(ILayoutedModelElement lobject, int index, int x, int y) {
		Points existingPoints = getConnectionPoints(lobject);
		Point newPoint = createPoint(x, y, null);
		existingPoints.getPoints().add(index, newPoint);
		setPoints(lobject, CONNECTION_POINTS, existingPoints);
	}

	/** Changes the point of the layout data at the given index. */
	public static void changeConnectionPoint(ILayoutedModelElement lobject, int index, int x,
			int y) {
		Points pts = getConnectionPoints(lobject);
		Point np = createPoint(x, y, null);
		pts.getPoints().set(index, np);
		setPoints(lobject, CONNECTION_POINTS, pts);
	}

	/** Changes all the point layout data of the given element. */
	public static void changeAllConnectionPoints(ILayoutedModelElement lobject, int dx, int dy) {
		Points pts = getConnectionPoints(lobject);
		Points newPoints = createPoints(null);
		if(pts != null) {
			for(Point p : pts.getPoints()) {
				Point np = createPoint(p.getX() + dx, p.getY() + dy, null);
				newPoints.getPoints().add(np);
			}
			setPoints(lobject, CONNECTION_POINTS, newPoints);
		}
	}

	/**
	 * Removes the point at the given index from the layout data of the given
	 * element.
	 */
	public static void removeConnectionPoint(ILayoutedModelElement lobject, int index) {
		Points pts = getConnectionPoints(lobject);
		pts.getPoints().remove(index);
		setPoints(lobject, CONNECTION_POINTS, pts);
	}

	/**
	 * Adds a {@link org.eclipse.draw2d.Bendpoint Bendpoint} defined by the x,y
	 * parameters to a channel
	 * 
	 * @param conn
	 *            the connection
	 * @param bendPointX
	 *            the x coordinate of the Bendpoint
	 * @param bendPointY
	 *            the y coordinate of the Bendpoint
	 */
	public static void addBendPointToConnection(ILayoutedModelElement conn, int bendPointX,
			int bendPointY) {
		// set middle point of transition
		Point bendPoint = LayoutFactory.eINSTANCE.createPoint();
		bendPoint.setX(bendPointX);
		bendPoint.setY(bendPointY);

		getConnectionPoints(conn).getPoints().add(bendPoint);
	}
}
